import math
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any

import cv2
import cv2.typing as cvt
import numpy as np

from homr import constants
from homr.image_utils import crop_image
from homr.simple_logging import eprint
from homr.type_definitions import NDArray


def calculate_edges_of_rotated_rectangle(
    box: cvt.RotatedRect,
) -> tuple[tuple[float, float], tuple[float, float], tuple[float, float], tuple[float, float]]:
    half_size = np.array([box[1][0] / 2, box[1][1] / 2])
    center = box[0]
    top_left = center - half_size
    bottom_left = center + np.array([-half_size[0], half_size[1]])
    top_right = center + np.array([half_size[0], -half_size[1]])
    bottom_right = center + half_size
    return (
        (top_left[0], top_left[1]),
        (bottom_left[0], bottom_left[1]),
        (top_right[0], top_right[1]),
        (bottom_right[0], bottom_right[1]),
    )


def do_polygons_overlap(poly1: cvt.MatLike, poly2: cvt.MatLike) -> bool:
    # Check if any point of one ellipse is inside the other ellipse
    for point in poly1:
        if cv2.pointPolygonTest(poly2, (float(point[0]), float(point[1])), False) >= 0:  # type: ignore
            return True
    for point in poly2:
        if cv2.pointPolygonTest(poly1, (float(point[0]), float(point[1])), False) >= 0:  # type: ignore
            return True

    return False


class DebugDrawable(ABC):
    @abstractmethod
    def draw_onto_image(self, img: NDArray, color: tuple[int, int, int] = (0, 0, 255)) -> None:
        pass


class AnyPolygon(DebugDrawable):
    def __init__(self, polygon: Any):
        self.polygon = polygon


class BoundingBox(AnyPolygon):
    """
    A bounding box in the format of (x1, y1, x2, y2)
    """

    def __init__(self, box: cvt.Rect, contours: cvt.MatLike, debug_id: int = 0):
        self.debug_id = debug_id
        self.contours = contours
        self.box = box
        self.center = (box[0] + box[2]) / 2, (box[1] + box[3]) / 2
        self.size = (box[2] - box[0], box[3] - box[1])
        self.rotated_box = (self.center, self.size, 0)
        self.size = (box[2] - box[0], box[3] - box[1])
        super().__init__(cv2.boxPoints(self.rotated_box).astype(np.int64))

    def draw_onto_image(self, img: NDArray, color: tuple[int, int, int] = (0, 0, 255)) -> None:
        cv2.rectangle(
            img,
            (int(self.box[0]), int(self.box[1])),
            (int(self.box[2]), int(self.box[3])),
            color,
            2,
        )

    def extract(self, img: NDArray) -> NDArray:
        return crop_image(img, self.box[0], self.box[1], self.box[2] + 1, self.box[3] + 1)

    def blank_everything_outside_of_box(self, img: NDArray) -> NDArray:
        x1, y1, x2, y2 = self.box
        img = img.copy()
        white_background = np.full_like(img, 255, dtype=np.uint8)
        white_background[y1:y2, x1:x2] = img[y1:y2, x1:x2]
        return white_background

    def increase_size_in_each_dimension(
        self, increase: int, image_size: tuple[int, ...]
    ) -> "BoundingBox":
        return BoundingBox(
            (
                max(self.box[0] - increase, 0),
                max(self.box[1] - increase, 0),
                min(self.box[2] + increase, image_size[1]),
                min(self.box[3] + increase, image_size[0]),
            ),
            self.contours,
            self.debug_id,
        )


class AngledBoundingBox(AnyPolygon):
    def __init__(
        self, box: cvt.RotatedRect, contours: cvt.MatLike, polygon: Any, debug_id: int = 0
    ):
        super().__init__(polygon)
        self.debug_id = debug_id
        self.contours = contours
        angle = box[2]
        self.box: cvt.RotatedRect
        if angle > 135:
            angle = angle - 180
            self.box = ((box[0][0], box[0][1]), (box[1][0], box[1][1]), angle)
        elif angle < -135:
            angle = angle + 180
            self.box = ((box[0][0], box[0][1]), (box[1][0], box[1][1]), angle)
        elif angle > 45:
            angle = angle - 90
            self.box = ((box[0][0], box[0][1]), (box[1][1], box[1][0]), angle)
        elif angle < -45:
            angle = angle + 90
            self.box = ((box[0][0], box[0][1]), (box[1][1], box[1][0]), angle)
        else:
            self.box = ((box[0][0], box[0][1]), (box[1][0], box[1][1]), angle)
        self.center = self.box[0]
        self.size = self.box[1]
        self.angle = self.box[2]
        self.top_left, self.bottom_left, self.top_right, self.bottom_right = (
            calculate_edges_of_rotated_rectangle(self.box)
        )
        self.polygon = polygon

    def is_overlapping(self, other: AnyPolygon) -> bool:
        if not self._can_shapes_possibly_touch(other):
            return False
        return do_polygons_overlap(self.polygon, other.polygon)

    def is_overlapping_with_any(self, others: Sequence["AngledBoundingBox"]) -> bool:
        for other in others:
            if self.is_overlapping(other):
                return True
        return False

    def _can_shapes_possibly_touch(self, other: "AnyPolygon") -> bool:
        """
        A fast check if the two shapes can possibly touch. If this returns False,
        the two shapes do not touch.
        If this returns True, the two shapes might touch and further checks are necessary.
        """

        # Get the centers and major axes of the rectangles
        center1, axes1, _ = self.box
        center2: Sequence[float]
        axes2: Sequence[float]
        if isinstance(other, BoundingBox):
            center2, axes2, _ = (
                other.rotated_box
            )  # (variable) rotated_box: tuple[tuple[float, float], tuple[int, int], Literal[0]]
        elif isinstance(other, AngledBoundingBox):
            center2, axes2, _ = (
                other.box
            )  # (variable) box: tuple[tuple[float, float], tuple[int, int], float]
        else:
            raise ValueError(f"Unknown type {type(other)}")
        major_axis1 = max(axes1)
        major_axis2 = max(axes2)

        # Calculate the distance between the centers
        distance = ((center1[0] - center2[0]) ** 2 + (center1[1] - center2[1]) ** 2) ** 0.5

        # If the distance is greater than the sum of the major axes, the rectangles do not overlap
        if distance > major_axis1 + major_axis2:
            return False
        return True

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, AngledBoundingBox):
            return self.box == __value.box
        else:
            return False

    def __hash__(self) -> int:
        return hash(self.box)

    def __str__(self) -> str:
        return f"{self.box}"

    def __repr__(self) -> str:
        return str(self)

    @abstractmethod
    def draw_onto_image(self, img: NDArray, color: tuple[int, int, int] = (0, 0, 255)) -> None:
        pass


class RotatedBoundingBox(AngledBoundingBox):
    def __init__(self, box: cvt.RotatedRect, contours: cvt.MatLike, debug_id: int = 0):
        super().__init__(box, contours, cv2.boxPoints(box).astype(np.int64), debug_id)

    def draw_onto_image(self, img: NDArray, color: tuple[int, int, int] = (0, 0, 255)) -> None:
        box = cv2.boxPoints(self.box).astype(np.int64)
        cv2.drawContours(img, [box], 0, color, 2)

    def is_intersecting(self, other: "RotatedBoundingBox") -> bool:
        if not self._can_shapes_possibly_touch(other):
            return False
        # TODO: How is this different from is_overlapping?
        return cv2.rotatedRectangleIntersection(self.box, other.box)[0] != cv2.INTERSECT_NONE

    def ensure_min_dimension(self, min_width: int, min_height: int) -> "RotatedBoundingBox":
        return RotatedBoundingBox(
            (
                (self.box[0][0], self.box[0][1]),
                (max(self.box[1][0], min_width), max(self.box[1][1], min_height)),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )

    def make_box_thicker(self, thickness: int) -> "RotatedBoundingBox":
        if thickness <= 0:
            return self
        # We tried to move the center by int(thickness / 2), however this gave much worse results
        # for some examples
        # That's possibly a case of an ill defined function, but downstream code depends on the
        # behavior which we have today
        return RotatedBoundingBox(
            (
                (self.box[0][0], self.box[0][1]),
                (self.box[1][0] + thickness, self.box[1][1] + thickness),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )

    def move_to_x_horizontal_by(self, x_delta: int) -> "RotatedBoundingBox":
        new_x = self.center[0] + x_delta
        return RotatedBoundingBox(
            ((new_x, self.center[1]), self.box[1], self.box[2]), self.contours, self.debug_id
        )

    def make_box_taller(self, thickness: int) -> "RotatedBoundingBox":
        return RotatedBoundingBox(
            (
                (self.box[0][0], self.box[0][1]),
                (self.box[1][0], self.box[1][1] + thickness),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )

    def make_box_taller_keep_center(self, thickness: int) -> "RotatedBoundingBox":
        return RotatedBoundingBox(
            (
                (self.box[0][0], self.box[0][1] - thickness // 2),
                (self.box[1][0], self.box[1][1] + thickness),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )

    def get_center_extrapolated(self, x: float) -> float:
        return (x - self.box[0][0]) * np.tan(self.box[2] / 180 * np.pi) + self.box[0][1]

    def is_overlapping_extrapolated(self, other: "RotatedBoundingBox", unit_size: float) -> bool:
        # Pick left and right by x-coordinate of the first box corner
        if self.box[0][0] > other.box[0][0]:
            left, right = other, self
        else:
            left, right = self, other

        center_x = (left.center[0] + right.center[0]) * 0.5

        tolerance = constants.tolerance_for_staff_line_detection(unit_size)
        max_gap = constants.max_line_gap_size(unit_size)

        if (
            center_x - left.center[0] - (left.size[0] // 2) > max_gap
            or right.center[0] - center_x - (right.size[0] // 2) > max_gap
        ):
            return False

        # Compute extrapolated y-values, same as get_center_extrapolated but inlined
        left_angle = math.tan(left.box[2] * math.pi / 180.0)
        right_angle = math.tan(right.box[2] * math.pi / 180.0)

        left_y = (center_x - left.box[0][0]) * left_angle + left.box[0][1]
        right_y = (center_x - right.box[0][0]) * right_angle + right.box[0][1]

        return abs(left_y - right_y) <= tolerance

    def to_bounding_box(self) -> BoundingBox:
        return BoundingBox(
            (
                int(self.top_left[0]),
                int(self.top_left[1]),
                int(self.bottom_right[0]),
                int(self.bottom_right[1]),
            ),
            self.contours,
            self.debug_id,
        )


class BoundingEllipse(AngledBoundingBox):
    def __init__(self, box: cvt.RotatedRect, contours: cvt.MatLike, debug_id: int = 0):
        super().__init__(
            box,
            contours,
            cv2.ellipse2Poly(
                (int(box[0][0]), int(box[0][1])),
                (int(box[1][0] / 2), int(box[1][1] / 2)),
                int(box[2]),
                0,
                360,
                1,
            ),
            debug_id,
        )

    def draw_onto_image(self, img: NDArray, color: tuple[int, int, int] = (0, 0, 255)) -> None:
        cv2.ellipse(img, self.box, color=color, thickness=2)

    def make_box_thicker(self, thickness: int) -> "BoundingEllipse":
        return BoundingEllipse(
            (
                (self.box[0][0], self.box[0][1]),
                (self.box[1][0] + thickness, self.box[1][1] + thickness),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )

    def make_box_taller(self, thickness: int) -> "RotatedBoundingBox":
        return RotatedBoundingBox(
            (
                (self.box[0][0], self.box[0][1]),
                (self.box[1][0], self.box[1][1] + thickness),
                self.box[2],
            ),
            self.contours,
            self.debug_id,
        )


def _has_box_valid_size(box: cvt.RotatedRect) -> bool:
    return (
        not math.isnan(box[1][0]) and not math.isnan(box[1][1]) and box[1][0] > 0 and box[1][1] > 0
    )


def create_rotated_bounding_boxes(
    img: NDArray,
    skip_merging: bool = False,
    min_size: tuple[int, int] | None = None,
    max_size: tuple[int, int] | None = None,
    thicken_boxes: int | None = None,
) -> list[RotatedBoundingBox]:
    contours, _ = cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    boxes: list[RotatedBoundingBox] = []
    for i, countour in enumerate(contours):
        fitBox = cv2.minAreaRect(countour)
        if not _has_box_valid_size(fitBox):
            continue
        box = RotatedBoundingBox(fitBox, countour, debug_id=i)
        if min_size is not None and (box.size[0] < min_size[0] or box.size[1] < min_size[1]):
            continue
        if max_size is not None:
            if max_size[0] > 0 and box.size[0] > max_size[0]:
                continue
            if max_size[1] > 0 and box.size[1] > max_size[1]:
                continue
        boxes.append(box)
    if skip_merging:
        return boxes
    if thicken_boxes is not None:
        boxes = [box.make_box_thicker(thicken_boxes) for box in boxes]
    return _get_box_for_whole_group(merge_overlaying_bounding_boxes(boxes))


def create_rotated_bounding_box(contour: cvt.MatLike, debug_id: int) -> RotatedBoundingBox:
    box = cv2.minAreaRect(contour)
    return RotatedBoundingBox(box, contour, debug_id=debug_id)


def create_lines(
    img: NDArray,
    threshold: int = 100,
    min_line_length: int = 100,
    max_line_gap: int = 10,
    skip_merging: bool = False,
) -> list[RotatedBoundingBox]:
    lines = cv2.HoughLinesP(
        img, 1, np.pi / 180, threshold, minLineLength=min_line_length, maxLineGap=max_line_gap
    )
    boxes = []
    for i, line in enumerate(lines):
        x1, y1, x2, y2 = line[0]  # type: ignore
        contour = np.array([[x1, y1], [x2, y2]])
        box = cv2.minAreaRect(contour)
        if box[1][0] > box[1][1]:
            boxes.append(RotatedBoundingBox(box, contour, debug_id=i))
    if skip_merging:
        return boxes
    return _get_box_for_whole_group(merge_overlaying_bounding_boxes(boxes))


def create_bounding_ellipses(
    img: NDArray,
    skip_merging: bool = False,
    min_size: tuple[int, int] | None = None,
    max_size: tuple[int, int] | None = None,
) -> list[BoundingEllipse]:
    contours, _ = cv2.findContours(img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    boxes = []
    for i, countour in enumerate(contours):
        min_length_to_fit_ellipse = 5  # this is a requirement by opencv
        if len(countour) < min_length_to_fit_ellipse:
            continue
        fitBox = cv2.fitEllipse(countour)
        if not _has_box_valid_size(fitBox):
            continue
        box = BoundingEllipse(fitBox, countour, debug_id=i)
        if min_size is not None and (box.size[0] < min_size[0] or box.size[1] < min_size[1]):
            continue
        if max_size is not None and (box.size[0] > max_size[0] or box.size[1] > max_size[1]):
            continue
        boxes.append(box)
    if skip_merging:
        return boxes
    return _get_ellipse_for_whole_group(merge_overlaying_bounding_boxes(boxes))


def _do_groups_overlap(group1: list[AngledBoundingBox], group2: list[AngledBoundingBox]) -> bool:
    for box1 in group1:
        for box2 in group2:
            if box1.is_overlapping(box2):
                return True
    return False


def _merge_groups_recursive(
    groups: list[list[AngledBoundingBox]], step: int
) -> list[list[AngledBoundingBox]]:
    step_limit = 10
    if step > step_limit:
        eprint("Too many steps in _merge_groups_recursive, giving back current results")
        return groups
    number_of_changes = 0
    merged: list[list[AngledBoundingBox]] = []
    used_groups = set()
    for i, group in enumerate(groups):
        match_found = False
        if i in used_groups:
            continue
        for j in range(i + 1, len(groups)):
            if j in used_groups:
                continue
            other_group = groups[j]
            if _do_groups_overlap(group, other_group):
                merged.append(group + other_group)
                number_of_changes += 1
                used_groups.add(j)
                match_found = True
                break
        if not match_found:
            merged.append(group)

    if number_of_changes == 0:
        return merged
    else:
        return _merge_groups_recursive(merged, step + 1)


def _get_ellipse_for_whole_group(groups: list[list[AngledBoundingBox]]) -> list[BoundingEllipse]:
    result = []
    for group in groups:
        complete_contour = np.concatenate([box.contours for box in group])
        box = cv2.minAreaRect(complete_contour)
        result.append(BoundingEllipse(box, complete_contour))
    return result


def _get_box_for_whole_group(groups: list[list[AngledBoundingBox]]) -> list[RotatedBoundingBox]:
    result = []
    for group in groups:
        complete_contour = np.concatenate([box.contours for box in group])
        box = cv2.minAreaRect(complete_contour)
        result.append(RotatedBoundingBox(box, complete_contour))
    return result


class UnionFind:
    def __init__(self, n: int):
        self.parent: list[int] = list(range(n))
        self.rank = [0] * n

    def find(self, x: int) -> int:
        if self.parent[x] != x:
            self.parent[x] = self.find(self.parent[x])
        return self.parent[x]

    def union(self, x: int, y: int) -> None:
        rootX = self.find(x)
        rootY = self.find(y)

        if rootX != rootY:
            # Union by rank to keep the tree flat
            if self.rank[rootX] > self.rank[rootY]:
                self.parent[rootY] = rootX
            elif self.rank[rootX] < self.rank[rootY]:
                self.parent[rootX] = rootY
            else:
                self.parent[rootY] = rootX
                self.rank[rootX] += 1


def _merge_groups_optimized(groups: list[list[AngledBoundingBox]]) -> list[list[AngledBoundingBox]]:
    n = len(groups)
    uf = UnionFind(n)

    # Try to find overlaps and union groups that overlap
    for i in range(n):
        for j in range(i + 1, n):
            if _do_groups_overlap(groups[i], groups[j]):
                uf.union(i, j)

    # Create merged groups based on the union-find results
    merged_groups: dict[int, list[AngledBoundingBox]] = {}
    for i in range(n):
        root = uf.find(i)
        if root not in merged_groups:
            merged_groups[root] = []
        merged_groups[root].extend(groups[i])

    return list(merged_groups.values())


def merge_overlaying_bounding_boxes(
    boxes: Sequence[AngledBoundingBox],
) -> list[list[AngledBoundingBox]]:
    initial_groups: list[list[AngledBoundingBox]] = []
    for box in boxes:
        initial_groups.append([box])
    return _merge_groups_optimized(initial_groups)



import os

script_location = os.path.dirname(os.path.realpath(__file__))

model_name = "segnet_308-3296ccd40960f90ca6ab9c035cca945675d30a0f"

segnet_path_onnx = os.path.join(script_location, f"{model_name}.onnx")

segnet_path_onnx_fp16 = os.path.join(script_location, f"{model_name}_fp16.onnx")

segnet_path_torch = os.path.join(
    os.getcwd(),
    "training",
    "architecture",
    "segmentation",
    f"{model_name}.pth",
)

segnet_version = os.path.basename(segnet_path_onnx).split("_")[1]

segmentation_version = segnet_version



import numpy as np

from homr.type_definitions import NDArray


def softmax(z: NDArray, dim: int = -1) -> NDArray:
    """Computes softmax function with support for dimension parameter.

    Args:
        z: input array
        dim: dimension along which softmax will be computed (-1 for last dimension)

    Returns an array of outputs with the same shape as z.
    """
    # For numerical stability: subtract the maximum along the specified axis
    shiftz = z - np.max(z, axis=dim, keepdims=True)
    exps = np.exp(shiftz)
    return exps / np.sum(exps, axis=dim, keepdims=True)



import numpy as np
import onnxruntime as ort
from onnxruntime import OrtValue

from homr.simple_logging import eprint
from homr.transformer.configs import Config
from homr.type_definitions import NDArray


class Encoder:
    def __init__(self, config: Config) -> None:
        self.use_gpu = False
        if config.use_gpu_inference:
            try:
                self.encoder = ort.InferenceSession(
                    config.filepaths.encoder_path_fp16,
                    providers=[
                        (
                            "CUDAExecutionProvider",
                            {
                                "cudnn_conv_algo_search": "DEFAULT",
                            },
                        )
                    ],
                )
                self.fp16 = True
                self.use_gpu = True

            except Exception as ex:
                eprint(ex)
                eprint("Going on without GPU support")
                self.encoder = ort.InferenceSession(config.filepaths.encoder_path_fp16)
                self.fp16 = True

        else:
            self.encoder = ort.InferenceSession(config.filepaths.encoder_path)
            self.fp16 = False

        self.io_binding = self.encoder.io_binding()
        self.device_id = 0

        self.input_name = self.encoder.get_inputs()[0].name
        self.output_name = self.encoder.get_outputs()[0].name

    def generate(self, x: NDArray) -> list[OrtValue]:
        if self.fp16:
            self.io_binding.bind_cpu_input("input", x.astype(np.float16))
        else:
            self.io_binding.bind_cpu_input("input", x.astype(np.float32))

        self.io_binding.bind_output("output", "cuda" if self.use_gpu else "cpu", self.device_id)
        self.encoder.run_with_iobinding(self.io_binding)
        return self.io_binding.get_outputs()[0].numpy()



import cv2

from homr.type_definitions import NDArray


def apply_clahe(image: NDArray) -> NDArray:
    image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    clahe = cv2.createCLAHE(clipLimit=1.0, tileGridSize=(8, 8))
    return clahe.apply(image)



from typing import Any

import numpy as np
import onnxruntime as ort

from homr.simple_logging import eprint
from homr.transformer.configs import Config
from homr.transformer.vocabulary import EncodedSymbol
from homr.type_definitions import NDArray


class ScoreDecoder:
    def __init__(
        self,
        transformer: ort.InferenceSession,
        fp16: bool,
        use_gpu: bool,
        config: Config,
        ignore_index: int = -100,
    ):
        super().__init__()
        self.ignore_index = ignore_index
        self.config = config
        self.net = transformer
        self.io_binding = self.net.io_binding()
        self.max_seq_len = config.max_seq_len
        self.eos_token = config.eos_token

        self.inv_rhythm_vocab = {v: k for k, v in config.rhythm_vocab.items()}
        self.inv_pitch_vocab = {v: k for k, v in config.pitch_vocab.items()}
        self.inv_lift_vocab = {v: k for k, v in config.lift_vocab.items()}
        self.inv_articulation_vocab = {v: k for k, v in config.articulation_vocab.items()}
        self.inv_position_vocab = {v: k for k, v in config.position_vocab.items()}

        self.fp16 = fp16
        self.use_gpu = use_gpu
        self.device_id = 0
        self.output_names = [
            "out_rhythms",
            "out_pitchs",
            "out_lifts",
            "out_positions",
            "out_articulations",
            "attention",
        ]

    def generate(
        self,
        start_tokens: NDArray,
        nonote_tokens: NDArray,
        **kwargs: Any,
    ) -> list[EncodedSymbol]:
        num_dims = len(start_tokens.shape)

        if num_dims == 1:
            start_tokens = start_tokens[None, :]

        b, t = start_tokens.shape

        out_rhythm = start_tokens
        out_pitch = nonote_tokens
        out_lift = nonote_tokens
        out_articulations = nonote_tokens
        cache, kv_input_names, kv_output_names = self.init_cache()
        output_names = self.output_names + kv_output_names
        context = kwargs["context"]
        context_reduced = kwargs["context"][:, :1]

        symbols: list[EncodedSymbol] = []

        for step in range(self.max_seq_len):
            x_lift = out_lift[:, -1:]  # for all: shape=(1,1)
            x_pitch = out_pitch[:, -1:]
            x_rhythm = out_rhythm[:, -1:]
            x_articulations = out_articulations[:, -1:]

            # after the first step we don't pass the full context into the decoder
            # x_transformers uses [:, :0] to split the context
            # which caused a Reshape error when loading the onnx model
            context = context if step == 0 else context_reduced

            # Bind Inputs
            self.io_binding.bind_cpu_input("rhythms", x_rhythm)
            self.io_binding.bind_cpu_input("pitchs", x_pitch)
            self.io_binding.bind_cpu_input("lifts", x_lift)
            self.io_binding.bind_cpu_input("articulations", x_articulations)
            self.io_binding.bind_cpu_input("context", context)
            self.io_binding.bind_cpu_input("cache_len", np.array([step], dtype=np.int64))
            for name, cache_val in zip(kv_input_names, cache, strict=True):
                self.io_binding.bind_ortvalue_input(name, cache_val)

            # Bind Outputs
            for name in output_names:
                self.io_binding.bind_output(name, "cuda" if self.use_gpu else "cpu", self.device_id)

            # Run inference
            self.net.run_with_iobinding(iobinding=self.io_binding)

            # Get outputs
            outputs = self.io_binding.get_outputs()
            cache = outputs[6:]

            # Greedy decoding: pick the highest logit directly for each output
            rhythmsp = outputs[0].numpy()
            pitchsp = outputs[1].numpy()
            liftsp = outputs[2].numpy()
            positionsp = outputs[3].numpy()
            articulationsp = outputs[4].numpy()
            attention = outputs[5].numpy()

            rhythm_sample = np.array([[rhythmsp[:, -1, :].argmax()]])
            pitch_sample = np.array([[pitchsp[:, -1, :].argmax()]])
            lift_sample = np.array([[liftsp[:, -1, :].argmax()]])
            articulation_sample = np.array([[articulationsp[:, -1, :].argmax()]])
            position_sample = np.array([[positionsp[:, -1, :].argmax()]])

            lift_token = detokenize(lift_sample, self.inv_lift_vocab)
            pitch_token = detokenize(pitch_sample, self.inv_pitch_vocab)
            rhythm_token = detokenize(rhythm_sample, self.inv_rhythm_vocab)
            articulation_token = detokenize(articulation_sample, self.inv_articulation_vocab)
            position_token = detokenize(position_sample, self.inv_position_vocab)

            if rhythm_sample[0][0] == self.eos_token:
                break

            symbol = EncodedSymbol(
                rhythm=rhythm_token[0],
                pitch=pitch_token[0],
                lift=lift_token[0],
                articulation=articulation_token[0],
                position=position_token[0],
                coordinates=attention,
            )
            symbols.append(symbol)

            out_lift = np.concatenate((out_lift, lift_sample), axis=-1)
            out_pitch = np.concatenate((out_pitch, pitch_sample), axis=-1)
            out_rhythm = np.concatenate((out_rhythm, rhythm_sample), axis=-1)
            out_articulations = np.concatenate((out_articulations, articulation_sample), axis=-1)

        return symbols

    def init_cache(self, cache_len: int = 0) -> tuple[list[NDArray], list[str], list[str]]:
        cache = []
        input_names = []
        output_names = []
        heads = self.config.decoder_heads
        head_dim = self.config.decoder_dim // heads
        for i in range(self.config.decoder_depth * 4):
            if self.fp16:  # the cache needs to be fp16 as well
                cache.append(
                    ort.OrtValue.ortvalue_from_numpy(
                        np.zeros((1, heads, cache_len, head_dim), dtype=np.float16),
                        "cuda" if self.use_gpu else "cpu",
                        self.device_id,
                    )
                )
            else:
                cache.append(
                    ort.OrtValue.ortvalue_from_numpy(
                        np.zeros((1, heads, cache_len, head_dim), dtype=np.float32),
                        "cuda" if self.use_gpu else "cpu",
                        self.device_id,
                    )
                )
            input_names.append(f"cache_in{i}")
            output_names.append(f"cache_out{i}")
        return cache, input_names, output_names


def detokenize(tokens: NDArray, vocab: dict[int, str]) -> list[str]:
    toks = [vocab[tok.item()] for tok in tokens]
    toks = [t for t in toks if t not in ("[BOS]", "[EOS]", "[PAD]")]
    return toks


def get_decoder(config: Config) -> ScoreDecoder:
    """
    Returns Tromr's Decoder
    """
    use_gpu = False
    if config.use_gpu_inference:
        try:
            onnx_transformer = ort.InferenceSession(
                config.filepaths.decoder_path_fp16, providers=["CUDAExecutionProvider"]
            )
            fp16 = True
            use_gpu = True
        except Exception as ex:
            eprint(ex)
            eprint("Going on without GPU support")
            onnx_transformer = ort.InferenceSession(config.filepaths.decoder_path_fp16)
            fp16 = True

    else:
        onnx_transformer = ort.InferenceSession(config.filepaths.decoder_path)
        fp16 = False

    return ScoreDecoder(onnx_transformer, fp16, use_gpu, config=config)



import cv2
import numpy as np

from homr import constants
from homr.bounding_boxes import RotatedBoundingBox
from homr.debug import Debug
from homr.model import MultiStaff, Staff
from homr.type_definitions import NDArray


def prepare_brace_dot_image(symbols: NDArray, staff: NDArray) -> NDArray:
    brace_dot = cv2.subtract(symbols, staff)
    """
    Remove horizontal lines and Make elements larger.
    """
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (1, 5))
    out = cv2.erode(brace_dot.astype(np.uint8), kernel)
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 35))
    return cv2.dilate(out, kernel)


def _filter_for_tall_elements(
    brace_dot: list[RotatedBoundingBox], staffs: list[Staff]
) -> list[RotatedBoundingBox]:
    """
    We filter elements in two steps:
    1. Use a rough unit size estimate to reduce the data size
    2. Find the closest staff and take its unit size to take warping into account
    """
    rough_unit_size = staffs[0].average_unit_size
    symbols_larger_than_rough_estimate = [
        symbol
        for symbol in brace_dot
        if symbol.size[1] > constants.min_height_for_brace_rough(rough_unit_size)
        and symbol.size[0] < constants.max_width_for_brace_rough(rough_unit_size)
    ]
    result = []
    for symbol in symbols_larger_than_rough_estimate:
        closest_staff = min(staffs, key=lambda staff: staff.y_distance_to(symbol.center))
        unit_size = closest_staff.average_unit_size
        if symbol.size[1] > constants.min_height_for_brace(unit_size):
            result.append(symbol)
    return result


def _get_connections_between_staffs_at_bar_lines(
    staff1: Staff, staff2: Staff, brace_dot: list[RotatedBoundingBox]
) -> list[RotatedBoundingBox]:
    bar_lines1 = staff1.get_bar_lines()
    bar_lines2 = staff2.get_bar_lines()
    result: list[RotatedBoundingBox] = []
    for symbol in brace_dot:
        symbol_thicker = symbol.make_box_thicker(30)
        first_overlapping_staff1 = [
            line for line in bar_lines1 if symbol_thicker.is_overlapping(line.box)
        ]
        first_overlapping_staff2 = [
            line for line in bar_lines2 if symbol_thicker.is_overlapping(line.box)
        ]
        if len(first_overlapping_staff1) >= 1 and len(first_overlapping_staff2) >= 1:
            result.append(symbol)
    return result


def _get_connections_between_staffs_at_clefs(
    staff1: Staff, staff2: Staff, brace_dot: list[RotatedBoundingBox]
) -> list[RotatedBoundingBox]:
    clefs1 = staff1.get_clefs()
    clefs2 = staff2.get_clefs()
    result: list[RotatedBoundingBox] = []
    for symbol in brace_dot:
        symbol_thicker = symbol.make_box_thicker(
            constants.tolerance_for_staff_at_any_point(staff1.average_unit_size)
        )
        first_overlapping_staff1 = [
            clef for clef in clefs1 if symbol_thicker.is_overlapping(clef.box)
        ]
        first_overlapping_staff2 = [
            clef for clef in clefs2 if symbol_thicker.is_overlapping(clef.box)
        ]
        if len(first_overlapping_staff1) >= 1 and len(first_overlapping_staff2) >= 1:
            result.append(symbol)
    return result


def _get_connections_between_staffs_at_lines(
    staff1: Staff, staff2: Staff, brace_dot: list[RotatedBoundingBox]
) -> list[RotatedBoundingBox]:
    result: list[RotatedBoundingBox] = []
    for symbol in brace_dot:
        symbol_thicker = symbol.make_box_thicker(
            constants.tolerance_for_touching_clefs(staff1.average_unit_size)
        )
        point1 = staff1.get_at(symbol.center[0])
        point2 = staff2.get_at(symbol.center[0])
        if point1 is None or point2 is None:
            continue
        if symbol_thicker.is_overlapping(
            point1.to_bounding_box()
        ) and symbol_thicker.is_overlapping(point2.to_bounding_box()):
            result.append(symbol)

    return result


def _get_connections_between_staffs(
    staff1: Staff, staff2: Staff, brace_dot: list[RotatedBoundingBox]
) -> list[RotatedBoundingBox]:
    result = []
    result.extend(_get_connections_between_staffs_at_bar_lines(staff1, staff2, brace_dot))
    result.extend(_get_connections_between_staffs_at_clefs(staff1, staff2, brace_dot))
    result.extend(_get_connections_between_staffs_at_lines(staff1, staff2, brace_dot))
    return result


def _merge_multi_staff_if_they_share_a_staff(staffs: list[MultiStaff]) -> list[MultiStaff]:
    """
    If two MultiStaff objects share a staff, merge them into one MultiStaff object.
    """
    result: list[MultiStaff] = []
    for staff in staffs:
        any_merged = False
        for existing in result:
            if len(set(staff.staffs).intersection(set(existing.staffs))) > 0:
                result.remove(existing)
                result.append(existing.merge(staff))
                any_merged = True
                break
        if not any_merged:
            result.append(staff)
    return result


def _create_grandstaffs(staffs: list[MultiStaff]) -> list[MultiStaff]:
    if len(staffs) == 0:
        return staffs
    return [s.create_grandstaffs() for s in staffs]


def find_braces_brackets_and_grand_staff_lines(
    debug: Debug, staffs: list[Staff], brace_dot: list[RotatedBoundingBox]
) -> list[MultiStaff]:
    """
    Connect staffs from multiple voices or grand staffs by searching for brackets and grand staffs.
    """
    brace_dot = _filter_for_tall_elements(brace_dot, staffs)
    result = []
    for i, staff in enumerate(staffs):
        neighbors: list[Staff] = []
        if i > 0:
            neighbors.append(staffs[i - 1])
        if i < len(staffs) - 1:
            neighbors.append(staffs[i + 1])
        any_connected_neighbor = False
        for neighbor in neighbors:
            connections = _get_connections_between_staffs(staff, neighbor, brace_dot)
            if len(connections) >= constants.minimum_connections_to_form_combined_staff:
                result.append(MultiStaff([staff, neighbor], connections))
                any_connected_neighbor = True
        if not any_connected_neighbor:
            result.append(MultiStaff([staff], []))

    return _create_grandstaffs(_merge_multi_staff_if_they_share_a_staff(result))



import os
import tarfile
import zipfile

import requests

from homr.simple_logging import eprint


def download_file(url: str, filename: str) -> None:
    response = requests.get(url, stream=True, timeout=5)
    total = int(response.headers.get("content-length", 0))
    totalMb = round(total / 1024 / 1024)
    last_percent = -1
    complete = 100

    os.makedirs(os.path.dirname(filename), exist_ok=True)

    with open(filename, "wb") as f:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                progress = f.tell()
                progressMb = round(progress / 1024 / 1024)
                if total > 0:
                    progressPercent = complete * progress // total
                    if progressPercent != last_percent:
                        eprint(
                            f"\rDownloaded {progressMb} of {totalMb} MB ({progressPercent}%)",
                            end="",
                        )
                        last_percent = progressPercent
                else:
                    eprint(f"\rDownloaded {progressMb} MB", end="")
    if total > 0 and last_percent != complete:
        eprint(f"\rDownloaded {totalMb} of {totalMb} MB (100%)")
    else:
        eprint()  # Add newline after download progress


def unzip_file(filename: str, output_folder: str, flatten_root_entry: bool = False) -> None:
    with zipfile.ZipFile(filename, "r") as zip_ref:
        zip_contents = zip_ref.namelist()

        if flatten_root_entry:
            common_prefix = os.path.commonprefix(zip_contents)
            if common_prefix and common_prefix.endswith("/"):
                zip_contents_dict = {
                    file: os.path.relpath(file, common_prefix) for file in zip_contents
                }
            else:
                zip_contents_dict = {file: file for file in zip_contents}
        else:
            zip_contents_dict = {file: file for file in zip_contents}

        for original, member in zip_contents_dict.items():
            # Ensure file path is safe
            if os.path.isabs(member) or ".." in member:
                eprint(f"Skipping potentially unsafe file {member}")
                continue

            # Handle directories
            if original.endswith("/"):
                os.makedirs(os.path.join(output_folder, member), exist_ok=True)
                continue

            # Extract file
            source = zip_ref.open(original)
            target = open(os.path.join(output_folder, member), "wb")

            with source, target:
                while True:
                    chunk = source.read(1024)
                    if not chunk:
                        break
                    target.write(chunk)


def untar_file(filename: str, output_folder: str) -> None:
    with tarfile.open(filename, "r:gz") as tar:
        for member in tar.getmembers():
            # Ensure file path is safe
            if os.path.isabs(member.name) or ".." in member.name:
                eprint(f"Skipping potentially unsafe file {member.name}")
                continue

            # Handle directories
            if member.type == tarfile.DIRTYPE:
                os.makedirs(os.path.join(output_folder, member.name), exist_ok=True)
                continue

            # Extract file
            source = tar.extractfile(member)
            if source is None:
                continue
            target = open(os.path.join(output_folder, member.name), "wb")

            with source, target:
                while True:
                    chunk = source.read(1024)
                    if not chunk:
                        break
                    target.write(chunk)



import copy
import itertools
import random
import re
from fractions import Fraction
from typing import Iterable

from homr.simple_logging import eprint

nonote = "."
empty = "_"  # used for decorations on note, if there is no decoration


def build_dict(tokens: Iterable[str]) -> dict[str, int]:
    result = {}
    for i, t in enumerate(tokens):
        if t in result:
            raise ValueError("Duplicated entry for " + t)
        if not t.strip():
            raise ValueError("Tokens must not be a whitespace, this makes parsing easier")
        if t.strip() != t:
            raise ValueError("Tokens must not contain a whitespace, this makes parsing easier")
        if "&" in t:
            raise ValueError("& is reserved as alternatives for chords")
        result[t] = i
    return result


def build_rhythm() -> dict[str, int]:
    rhythm = []

    # sequence symbols
    rhythm.extend(["PAD", "BOS", "EOS"])
    rhythm.append("chord")

    # bar lines
    rhythm.extend(["barline", "doublebarline", "bolddoublebarline"])
    rhythm.extend(
        ["repeatStart", "repeatEnd", "repeatEndStart"]
    )  # , "daCapo", "daSegno", "segno", "coda"
    rhythm.extend(["voltaStart", "voltaStop", "voltaDiscontinue"])

    # clefs
    rhythm.extend([f"clef_F{c}" for c in range(3, 6)])
    rhythm.extend([f"clef_C{c}" for c in range(1, 6)])
    rhythm.extend([f"clef_G{c}" for c in range(1, 3)])

    # signatures
    rhythm.extend([f"keySignature_{c}" for c in range(-7, 8)])
    rhythm.extend([f"timeSignature/{c}" for c in [1, 2, 3, 4, 6, 8, 12, 16, 32, 48]])

    # rhythm, kern durations are based on https://www.humdrum.org/rep/kern/
    rhythm.extend([f"rest_{c}m" for c in range(2, 11)])  # multirests
    kern_base_durations = [0, 1, 2, 3, 4, 5, 6, 8, 10, 12, 16, 32, 64, 128]  # 128, 256, 512
    dots = ["", ".", ".."]
    grace = ["", "G"]
    kern_values = [
        f"{d}{g}{dot}" for d, g, dot in itertools.product(kern_base_durations, grace, dots)
    ]

    # Durations coming from tuplets
    irregular_durations = [7, 11, 13, 18, 20, 21, 22, 24, 26, 28, 30, 34, 36, 40, 48, 56, 96]

    rhythm.extend([f"note_{d}" for d in kern_values])
    rhythm.extend([f"note_{d}" for d in irregular_durations])
    rhythm.extend([f"rest_{d}" for d in kern_values])
    rhythm.extend([f"rest_{d}" for d in irregular_durations])

    # Note relations (https://en.wikipedia.org/wiki/List_of_musical_symbols)
    rhythm.extend(["tieSlur"])  #  "gliss"

    # Dynamics
    # rhythm.extend(
    #    [f"dynamic_{d}" for d in ["ppp", "pp", "p", "mp", "mf", "f", "ff", "fff", "sfz", "fp"]]
    # )
    # rhythm.extend(
    #    [
    #        "crescendoStart",
    #        "crescendoEnd",
    #        "diminuendoStart",
    #        "diminuendoEnd",
    #    ]
    # )

    return build_dict(rhythm)


def build_lift() -> dict[str, int]:
    lifts = [nonote, empty, "#", "##", "N", "b", "bb"]
    return build_dict(lifts)


def build_position() -> dict[str, int]:
    """
    The staff position, applies to notes, rests and clefs
    """
    positions = [nonote, "upper", "lower"]
    return build_dict(positions)


def build_articulation() -> dict[str, int]:
    articulation = [nonote, empty]

    # The lieder dataset has the superset of articulations
    # so we use it to define this list
    articulations_lieder = [
        "accent",
        "accent_arpeggiate",
        "accent_arpeggiate_fermata",
        "accent_arpeggiate_slurStart",
        "accent_arpeggiate_slurStart_staccato",
        "accent_arpeggiate_slurStop",
        "accent_arpeggiate_slurStop_staccato",
        "accent_arpeggiate_staccato",
        "accent_breathMark",
        "accent_breathMark_fermata",
        "accent_breathMark_slurStart",
        "accent_breathMark_slurStop",
        "accent_fermata",
        "accent_fermata_slurStart",
        "accent_fermata_slurStart_trill",
        "accent_fermata_slurStop",
        "accent_slurStart",
        "accent_slurStart_slurStop",
        "accent_slurStart_slurStop_tenuto",
        "accent_slurStart_slurStop_trill",
        "accent_slurStart_staccato",
        "accent_slurStart_tenuto",
        "accent_slurStart_tieStart",
        "accent_slurStart_tieStop",
        "accent_slurStart_tremolo",
        "accent_slurStart_trill",
        "accent_slurStop",
        "accent_slurStop_staccato",
        "accent_slurStop_tenuto",
        "accent_slurStop_tieStart",
        "accent_slurStop_tremolo",
        "accent_slurStop_trill",
        "accent_staccatissimo",
        "accent_staccato",
        "accent_staccato_tenuto",
        "accent_tenuto",
        "accent_tieStart",
        "accent_tieStart_trill",
        "accent_tieStop",
        "accent_tremolo",
        "accent_trill",
        "arpeggiate",
        "arpeggiate_fermata",
        "arpeggiate_fermata_slurStart",
        "arpeggiate_fermata_slurStop",
        "arpeggiate_slurStart",
        "arpeggiate_slurStart_slurStop",
        "arpeggiate_slurStart_staccatissimo",
        "arpeggiate_slurStart_staccato",
        "arpeggiate_slurStart_staccato_tenuto",
        "arpeggiate_slurStart_tenuto",
        "arpeggiate_slurStart_tremolo",
        "arpeggiate_slurStop",
        "arpeggiate_slurStop_staccatissimo",
        "arpeggiate_slurStop_staccato",
        "arpeggiate_slurStop_tenuto",
        "arpeggiate_staccatissimo",
        "arpeggiate_staccato",
        "arpeggiate_tenuto",
        "arpeggiate_tremolo",
        "arpeggiate_trill",
        "breathMark",
        "breathMark_fermata",
        "breathMark_fermata_slurStop",
        "breathMark_fermata_tenuto",
        "breathMark_slurStart",
        "breathMark_slurStart_trill",
        "breathMark_slurStop",
        "breathMark_slurStop_staccato",
        "breathMark_slurStop_trill",
        "breathMark_staccato",
        "breathMark_staccato_tenuto",
        "breathMark_tenuto",
        "breathMark_tremolo",
        "fermata",
        "fermata_slurStart",
        "fermata_slurStart_slurStop",
        "fermata_slurStart_trill",
        "fermata_slurStop",
        "fermata_slurStop_staccato",
        "fermata_slurStop_tenuto",
        "fermata_slurStop_tremolo",
        "fermata_slurStop_trill",
        "fermata_slurStop_turn",
        "fermata_staccato",
        "fermata_tenuto",
        "fermata_tieStart_tieStop",
        "fermata_tieStop",
        "fermata_tremolo",
        "fermata_trill",
        "slurStart",
        "slurStart_slurStop",
        "slurStart_slurStop_staccato",
        "slurStart_slurStop_staccato_tenuto",
        "slurStart_slurStop_tenuto",
        "slurStart_slurStop_tieStart_tieStop",
        "slurStart_slurStop_tremolo",
        "slurStart_slurStop_trill",
        "slurStart_slurStop_turn",
        "slurStart_staccatissimo",
        "slurStart_staccato",
        "slurStart_staccato_tenuto",
        "slurStart_staccato_tieStop",
        "slurStart_tenuto",
        "slurStart_tenuto_tieStart",
        "slurStart_tenuto_trill",
        "slurStart_tieStart",
        "slurStart_tieStart_tieStop",
        "slurStart_tieStop",
        "slurStart_tieStop_trill",
        "slurStart_tremolo",
        "slurStart_trill",
        "slurStart_turn",
        "slurStop",
        "slurStop_staccatissimo",
        "slurStop_staccato",
        "slurStop_staccato_tenuto",
        "slurStop_tenuto",
        "slurStop_tieStart",
        "slurStop_tieStart_tieStop",
        "slurStop_tieStop",
        "slurStop_tremolo",
        "slurStop_trill",
        "slurStop_turn",
        "staccatissimo",
        "staccato",
        "staccato_tenuto",
        "staccato_tieStart",
        "staccato_tieStop",
        "staccato_tremolo",
        "staccato_trill",
        "staccato_turn",
        "tenuto",
        "tieStart",
        "tieStart_tieStop",
        "tieStart_tieStop_trill",
        "tieStart_trill",
        "tieStop",
        "tieStop_trill",
        "tremolo",
        "trill",
        "turn",
    ]

    articulation.extend(articulations_lieder)

    return build_dict(articulation)


def build_pitch() -> dict[str, int]:
    pitch = [nonote, empty]
    note_names = ["C", "D", "E", "F", "G", "A", "B"]
    octave = range(10)
    pitch.extend(reversed([f"{n}{octave}" for octave, n in itertools.product(octave, note_names)]))
    return build_dict(pitch)


def has_rhythm_symbol_a_position(rhythm: str) -> bool:
    return rhythm.startswith(("note", "rest", "clef"))


class Vocabulary:
    def __init__(self) -> None:
        self.rhythm = build_rhythm()
        self.lift = build_lift()
        self.articulation = build_articulation()
        self.pitch = build_pitch()
        self.position = build_position()


class SymbolDuration:
    def __init__(
        self, base_duration: Fraction, dots: int, actual_notes: int, normal_notes: int, kern: int
    ) -> None:
        self.base_duration = base_duration
        self.dots = dots
        action_normal = Fraction(actual_notes, normal_notes)
        self.actual_notes = action_normal.numerator
        self.normal_notes = action_normal.denominator
        self.fraction = self._to_fraction()
        self.kern = kern

    def _to_fraction(self) -> Fraction:
        """
        Convert to a final Fraction duration (relative to whole note).
        """
        dur = self.base_duration

        # Apply dots
        add = dur / 2
        for _ in range(self.dots):
            dur += add
            add /= 2

        # Apply tuplet scaling
        if self.actual_notes != self.normal_notes:
            dur *= Fraction(self.normal_notes, self.actual_notes)

        return dur


def prior_power_of_two(n: int) -> int:
    """Return the next power of two <= n."""
    if n < 1:
        # This produces wrong rhythms, but at least it produces something
        return 1
    return 1 << (n.bit_length() - 1)


def kern_to_symbol_duration(kern: str) -> SymbolDuration:
    """
    Parse a Humdrum **kern duration string into a SymbolDuration object.
    """
    if kern.endswith("m"):
        # Multirest
        SymbolDuration(Fraction(1), 0, 1, 1, 4)

    # Extract numeric prefix (can be > 1 digit)
    i = 0
    while i < len(kern) and kern[i].isdigit():
        i += 1
    base_str = kern[:i]
    rest = kern[i:]

    base = int(base_str) if base_str else 4  # default quarter
    dots = rest.count(".")

    if "G" in kern:
        # Grace note
        return SymbolDuration(Fraction(0), dots, 1, 1, base)
    if base == 0:
        # Special: Whoe measure rest
        return SymbolDuration(Fraction(1), dots, 1, 1, base)

    # If base is a power of two, it's a normal note
    if base & (base - 1) == 0:
        base_duration = Fraction(1, base)
        actual_notes, normal_notes = 1, 1
        return SymbolDuration(base_duration, dots, actual_notes, normal_notes, base)
    else:
        # Tuplet case: find next higher power of two
        normal_notes = prior_power_of_two(base)
        base_duration = Fraction(1, normal_notes)
        actual_notes = base
        return SymbolDuration(base_duration, dots, actual_notes, normal_notes, normal_notes)


class EncodedSymbol:
    """
    A musical symbol split into the different decoder branches.
    """

    def __init__(
        self,
        rhythm: str,
        pitch: str = nonote,
        lift: str = nonote,
        articulation: str = nonote,
        position: str = nonote,
        coordinates: tuple[float, float] | None = None,
    ) -> None:
        self.rhythm = rhythm
        self.pitch = pitch
        self.lift = lift
        self.articulation = articulation
        self.position = position

        # These coordinates are derived from transformer attention and are inherently imprecise,
        # since the model is optimized for predictive accuracy rather than spatial localization.
        # Because patch tokens are processed in raster order (top-to-bottom, left-to-right),
        # this ordering can be used to reject cases where attention-based coordinates
        # violate monotonic scan constraints and are therefore unreliable.
        self.coordinates = coordinates
        self._duration: SymbolDuration | None = None

    def is_control_symbol(self) -> bool:
        return self.rhythm in ("BOS", "EOS", "PAD")

    def is_tuplet(self) -> bool:
        no_tuplet = self.remove_tuplet()
        return no_tuplet.rhythm != self.rhythm

    def remove_tuplet(self) -> "EncodedSymbol":
        match = re.match(r"(note|rest)_(\d+)(.*)", self.rhythm)
        if not match:
            return self
        duration = int(match[2])
        if duration % 3 == 0:
            duration = duration // 3 * 2
        elif duration % 5 == 0:
            duration = duration // 5 * 4
        elif duration % 7 == 0:
            duration = duration // 7 * 4
        else:
            return self

        result = copy.copy(self)
        result.rhythm = match[1] + "_" + str(duration) + match[3]
        result._duration = None
        return result

    def change_lift(self, lift: str) -> "EncodedSymbol":
        result = copy.copy(self)
        result.lift = lift
        return result

    def to_upper_position(self) -> "EncodedSymbol":
        if self.position != "lower":
            return self
        result = copy.copy(self)
        result.position = "upper"
        return result

    def is_valid(self) -> bool:
        has_position = has_rhythm_symbol_a_position(self.rhythm)
        is_note = [s != nonote for s in [self.lift, self.articulation, self.pitch, self.position]]
        return all(item == has_position for item in is_note)

    def add_articulations(self, articulations: list[str]) -> "EncodedSymbol":
        all_articulations = []
        all_articulations.extend(articulations)
        all_articulations.extend([a for a in self.articulation.split("_") if a])
        result = copy.copy(self)
        result.articulation = str.join("_", sorted(all_articulations))
        return result

    def strip_articulations(
        self, to_be_removed: list[str], remove_all: bool = False
    ) -> tuple[list[str], "EncodedSymbol"]:
        stripped = []
        remaining = []
        for articulation in self.articulation.split("_"):
            if not articulation:
                continue
            if remove_all or articulation in to_be_removed:
                stripped.append(articulation)
            else:
                remaining.append(articulation)
        result = copy.copy(self)
        if remaining:
            result.articulation = str.join("_", remaining)
        else:
            result.articulation = empty

        return stripped, result

    def get_duration(self) -> SymbolDuration:
        """
        Convert a Humdrum **kern duration into a Fraction (relative to a whole note).
        """
        if self._duration is not None:
            return self._duration

        rhythm = self.rhythm
        if not rhythm.startswith(("note", "rest")):
            eprint("Warning, invalid symbol in group: Only notes and rests have durations")
            return SymbolDuration(Fraction(0), 0, 1, 1, 1)
        kern = rhythm.split("_")[1]

        duration = kern_to_symbol_duration(kern)
        self._duration = duration
        return duration

    def __str__(self) -> str:
        return str.join(" ", (self.rhythm, self.pitch, self.lift, self.articulation, self.position))

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, EncodedSymbol):
            return (
                self.rhythm == __value.rhythm
                and self.pitch == __value.pitch
                and self.lift == __value.lift
                and self.articulation == __value.articulation
                and self.position == __value.position
            )
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.rhythm, self.pitch, self.lift, self.articulation, self.position))

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, EncodedSymbol):
            return NotImplemented
        return str(self) > str(other)


def _remove_redudant_clefs_keys_and_time_signatures(
    chords: list[list[EncodedSymbol]],
) -> list[list[EncodedSymbol]]:
    clef_upper = ""
    clef_lower = ""
    key = ""
    time = ""
    result_chords = []
    for chord in chords:
        result = []
        for symbol in chord:
            if symbol.rhythm.startswith("clef"):
                if symbol.position == "upper":
                    if symbol.rhythm != clef_upper:
                        clef_upper = symbol.rhythm
                        result.append(symbol)
                elif symbol.rhythm != clef_lower:
                    clef_lower = symbol.rhythm
                    result.append(symbol)
            elif symbol.rhythm.startswith("keySignature"):
                if symbol.rhythm != key:
                    key = symbol.rhythm
                    result.append(symbol)
            elif symbol.rhythm.startswith("timeSignature"):
                if symbol.rhythm != time:
                    time = symbol.rhythm
                    result.append(symbol)
            else:
                result.append(symbol)
        result_chords.append(result)
    return result_chords


def _remove_duplicated_piches(chord: list[EncodedSymbol]) -> list[EncodedSymbol]:
    if len(chord) <= 1 or not chord[0].rhythm.startswith(("note", "rest")):
        return chord
    by_pitch: dict[str, EncodedSymbol] = {}
    order_of_appearance = []
    for symbol in chord:
        key = symbol.pitch + " " + symbol.position
        if key in by_pitch:
            if symbol.get_duration().fraction > by_pitch[key].get_duration().fraction:
                by_pitch[symbol.pitch] = symbol
        else:
            by_pitch[key] = symbol
            order_of_appearance.append(key)

    return [by_pitch[s] for s in order_of_appearance]


def _group_into_chords(symbols: list[EncodedSymbol]) -> list[list[EncodedSymbol]]:
    chords: list[list[EncodedSymbol]] = []
    is_in_chord = False
    for symbol in symbols:
        if symbol.rhythm == "chord":
            is_in_chord = True
        elif is_in_chord and len(chords) > 0:
            chords[-1].append(symbol)
            is_in_chord = False
        else:
            chords.append([symbol])
    return chords


def _flatten_chords(chords: list[list[EncodedSymbol]]) -> list[EncodedSymbol]:
    result = []
    for chord in chords:
        if len(chords) == 0:
            continue
        is_in_chord = False
        for symbol in chord:
            if is_in_chord:
                result.append(EncodedSymbol("chord"))
            result.append(symbol)
            is_in_chord = True

    return result


def _group_into_measures(chords: list[list[EncodedSymbol]]) -> list[list[list[EncodedSymbol]]]:
    measures = []
    current_measure = []
    for chord in chords:
        current_measure.append(chord)
        if len(chord) > 0 and ("barline" in chord[0].rhythm or "repeat" in chord[0].rhythm):
            measures.append(current_measure)
            current_measure = []
    if len(current_measure) > 0:
        measures.append(current_measure)
    return measures


def _flatten_measures(measures: list[list[list[EncodedSymbol]]]) -> list[list[EncodedSymbol]]:
    return [chord for measure in measures for chord in measure]


def _get_duration_of_measure(measure: list[list[EncodedSymbol]]) -> Fraction:
    total_duration = Fraction(0)
    for chord in measure:
        duration = Fraction(0)
        for symbol in chord:
            if symbol.rhythm.startswith(("note", "rest")):
                fraction = symbol.get_duration().fraction
                if fraction > Fraction(0) and (fraction < duration or duration == Fraction(0)):
                    duration = fraction
        total_duration += duration
    return total_duration


def _get_typical_duration_of_measures(measures: list[list[list[EncodedSymbol]]]) -> Fraction:
    durations = [_get_duration_of_measure(m) for m in measures]
    if len(durations) == 0:
        return Fraction(0)
    return sorted(durations)[len(durations) // 2]


def _remove_tuplets(measure: list[list[EncodedSymbol]]) -> list[list[EncodedSymbol]]:
    return [[symbol.remove_tuplet() for symbol in chord] for chord in measure]


def _fix_over_eager_tuplets(chords: list[list[EncodedSymbol]]) -> list[list[EncodedSymbol]]:
    """
    The transformer tends to add too many tuplets, so we remove them
    based on the length of a measurement.
    """
    measures = _group_into_measures(chords)
    mean = _get_typical_duration_of_measures(measures)
    result = []
    for i, measure in enumerate(measures):
        if _get_duration_of_measure(measure) < mean:
            eprint("Removing tuplets from measure #", i + 1)
            result.append(_remove_tuplets(measure))
        else:
            result.append(measure)
    return _flatten_measures(result)


def _only_keep_lower_staff_if_there_is_a_clef(
    chords: list[list[EncodedSymbol]],
) -> list[list[EncodedSymbol]]:
    has_lower_clef = False
    all_results = []
    for i, chord in enumerate(chords):
        result = []
        for symbol in chord:
            if has_lower_clef:
                result.append(symbol)
            elif i < 5 and symbol.rhythm.startswith("clef") and symbol.position == "lower":
                has_lower_clef = True
                result.append(symbol)
            else:
                result.append(symbol.to_upper_position())
        all_results.append(result)
    delta = len(chords) - len(all_results)
    if delta > 0:
        eprint("Removed", delta, "results as there was no matching clef")
    return all_results


def remove_duplicated_symbols(
    symbols: list[EncodedSymbol], cleanup_tuplets: bool = True
) -> list[EncodedSymbol]:
    chords = _group_into_chords(symbols)
    if cleanup_tuplets:
        chords = _fix_over_eager_tuplets(chords)
        chords = _only_keep_lower_staff_if_there_is_a_clef(chords)
    chords = [_remove_duplicated_piches(c) for c in chords]
    chords = _remove_redudant_clefs_keys_and_time_signatures(chords)
    return _flatten_chords(chords)


def sort_token_chords(
    symbols: list[EncodedSymbol], keep_chord_symbol: bool = False
) -> list[list[EncodedSymbol]]:
    chords: list[list[EncodedSymbol]] = []
    is_in_chord = False
    for symbol in symbols:
        if symbol.rhythm == "chord":
            is_in_chord = True
        elif is_in_chord and len(chords) > 0:
            if keep_chord_symbol:
                chords[-1].append(EncodedSymbol("chord"))
            chords[-1].append(symbol)
            is_in_chord = False
        else:
            chords.append([symbol])

    return [sorted(chord) for chord in chords]


if __name__ == "__main__":
    import json

    from homr.simple_logging import eprint

    vocab = Vocabulary()

    eprint("Rhythm=", json.dumps(vocab.rhythm, indent=2))
    eprint("Lift=", json.dumps(vocab.lift, indent=2))
    eprint("Articulation=", json.dumps(vocab.articulation, indent=2))
    eprint("Pitch=", json.dumps(vocab.pitch, indent=2))
    eprint("Positions=", json.dumps(vocab.position, indent=2))

    valid_combinations = []

    for r, li, a, p, pos in itertools.product(
        vocab.rhythm, vocab.lift, vocab.articulation, vocab.pitch, vocab.position
    ):
        symbol = EncodedSymbol(r, li, a, p, pos)
        is_valid = symbol.is_valid()
        if not is_valid:
            continue
        valid_combinations.append(symbol)

    eprint(
        "Number of combinations",
        len(valid_combinations),
        "- some examples:",
    )
    for symbol in random.sample(valid_combinations, 10):
        eprint(symbol)



from homr.model import MultiStaff, Staff


def get_center_min_max_y(staff: Staff) -> tuple[float, float]:
    return (staff.min_y, staff.max_y)


class StaffRegions:
    def __init__(self, staffs: list[MultiStaff]) -> None:
        self.centers = [get_center_min_max_y(s) for ms in staffs for s in ms.staffs]

    def get_start_of_closest_staff_above(self, y: float) -> float:
        staffs_above = [c[1] for c in self.centers if c[0] < y]
        if len(staffs_above) == 0:
            return 0

        return max(staffs_above)

    def get_start_of_closest_staff_below(self, y: float) -> float:
        staffs_below = [c[0] for c in self.centers if c[1] > y]
        if len(staffs_below) == 0:
            return 1e12  # should be larger than the height of every reasonable image

        return min(staffs_below)



import abc
from abc import ABC

from homr.simple_logging import eprint
from homr.transformer.vocabulary import EncodedSymbol, empty, nonote

definition = {
    -7: "CbM",
    -6: "GbM",
    -5: "DbM",
    -4: "AbM",
    -3: "EbM",
    -2: "BbM",
    -1: "FM",
    0: "CM",
    1: "GM",
    2: "DM",
    3: "AM",
    4: "EM",
    5: "BM",
    6: "F#M",
    7: "C#M",
}

inv_definition = {v: k for k, v in definition.items()}

circle_of_fifth_notes_positive = ["F", "C", "G", "D", "A", "E", "B"]
circle_of_fifth_notes_negative = list(reversed(circle_of_fifth_notes_positive))


def key_signature_to_circle_of_fifth(key_signature: str) -> int:
    if key_signature not in inv_definition:
        eprint("Warning: Unknown key signature", key_signature)
        return 0
    return inv_definition[key_signature]


def repeat_note_for_all_octaves(notes: list[str]) -> list[str]:
    """
    Takes a list of notes and returns a list of notes that includes all octaves.
    """

    result = []

    for note in notes:
        for octave in range(11):
            result.append(note + str(octave))
    return result


class AbstractKeyTransformation(ABC):

    @abc.abstractmethod
    def add_accidental(self, note: str, accidental: str) -> str:
        pass

    @abc.abstractmethod
    def reset_at_end_of_measure(self) -> "AbstractKeyTransformation":
        pass


class NoKeyTransformation(AbstractKeyTransformation):

    def __init__(self) -> None:
        self.current_accidentals: dict[str, str] = {}

    def add_accidental(self, note: str, accidental: str) -> str:
        if accidental != "" and (
            note not in self.current_accidentals or self.current_accidentals[note] != accidental
        ):
            self.current_accidentals[note] = accidental
            return accidental
        else:
            return ""

    def reset_at_end_of_measure(self) -> "NoKeyTransformation":
        return NoKeyTransformation()


class KeyTransformation(AbstractKeyTransformation):

    def __init__(self, circle_of_fifth: int):
        self.circle_of_fifth = circle_of_fifth
        self.sharps: set[str] = set()
        self.flats: set[str] = set()
        if circle_of_fifth > 0:
            self.sharps = set(
                repeat_note_for_all_octaves(circle_of_fifth_notes_positive[0:circle_of_fifth])
            )
        elif circle_of_fifth < 0:
            self.flats = set(
                repeat_note_for_all_octaves(
                    circle_of_fifth_notes_negative[0 : abs(circle_of_fifth)]
                )
            )

    def add_accidental(self, note: str, accidental: str | None) -> str:
        """
        Returns the accidental if it wasn't placed before.
        """

        if accidental in ["#", "b", "N"]:
            previous_accidental = "N"
            if note in self.sharps:
                self.sharps.remove(note)
                previous_accidental = "#"
            if note in self.flats:
                self.flats.remove(note)
                previous_accidental = "b"
            if accidental == "#":
                self.sharps.add(note)
            elif accidental == "b":
                self.flats.add(note)
            return accidental if accidental != previous_accidental else ""
        else:
            if note in self.sharps:
                self.sharps.remove(note)
                return "N"

            if note in self.flats:
                self.flats.remove(note)
                return "N"
            return ""

    def reset_at_end_of_measure(self) -> "KeyTransformation":
        return KeyTransformation(self.circle_of_fifth)


def maintain_accidentals_during_measure(
    symbols: list[EncodedSymbol],
) -> list[EncodedSymbol]:
    """
    The PrIMuS datset doesn't maintain accidentals. Possibly
    because it uses a different rule for accidentals as in music
    there is no aggreement on this matter. However homr and the
    other datasets maintain accidentals until the endo f the measure,
    so we adjust the PrIMus ground truth to match this.

    Example: datasets/Corpus/000135772-1_2_1/000135772-1_2_1-pre.jpg
    """
    results = []

    # Since PrIMuS treats keys as we expect, we don't
    # have to take the key into account and therefore just the
    # key of C as it has no accidentals
    key = KeyTransformation(0)

    for symbol in symbols:
        if "barline" in symbol.rhythm:
            key = key.reset_at_end_of_measure()
            results.append(symbol)
        elif symbol.lift != nonote:
            # In engraving, the lift may be empty (implied by key signature or previous accidental)
            # In sounding, we need the actual pitch: remove any previous
            # accidental tracking to force the current one
            lift = symbol.lift if symbol.lift != empty else None
            actual_accidental = None

            if lift in ["#", "b", "N"]:
                actual_accidental = lift
                # Update key state to reflect that this accidental has been
                # applied for future notes in the measure
                key.add_accidental(symbol.pitch, lift)
            elif symbol.pitch in key.sharps:  # Determine if note is sharp/flat by key signature
                actual_accidental = "#"
            elif symbol.pitch in key.flats:
                actual_accidental = "b"
            else:
                actual_accidental = empty

            results.append(symbol.change_lift(actual_accidental if actual_accidental else empty))
        else:
            results.append(symbol)

    return results


def strip_naturals(
    symbols: list[EncodedSymbol],
) -> list[EncodedSymbol]:
    results = []
    for symbol in symbols:
        if symbol.lift == "N":
            results.append(symbol.change_lift(empty))
        else:
            results.append(symbol)

    return results



import math

import cv2
import numpy as np

from homr import constants
from homr.debug import Debug
from homr.image_utils import crop_image_and_return_new_top
from homr.model import MultiStaff, Staff
from homr.simple_logging import eprint
from homr.staff_dewarping import StaffDewarping, dewarp_staff_image
from homr.staff_parsing_tromr import parse_staff_tromr
from homr.staff_regions import StaffRegions
from homr.transformer.configs import Config, default_config
from homr.transformer.vocabulary import EncodedSymbol, remove_duplicated_symbols
from homr.type_definitions import NDArray


def _have_all_the_same_number_of_staffs(staffs: list[MultiStaff]) -> bool:
    for staff in staffs:
        if len(staff.staffs) != len(staffs[0].staffs):
            return False
    return True


def _is_close_to_image_top_or_bottom(staff: MultiStaff, image: NDArray) -> bool:
    tolerance = 50.0
    closest_distance_to_top_or_bottom: list[float] = [
        min(s.min_x, image.shape[0] - s.max_x) for s in staff.staffs
    ]
    return min(closest_distance_to_top_or_bottom) < tolerance


def _ensure_same_number_of_staffs(staffs: list[MultiStaff], image: NDArray) -> list[MultiStaff]:
    if _have_all_the_same_number_of_staffs(staffs):
        return staffs
    if len(staffs) > 2:
        if _is_close_to_image_top_or_bottom(
            staffs[0], image
        ) and _have_all_the_same_number_of_staffs(staffs[1:]):
            eprint("Removing first system from all voices, as it has a different number of staffs")
            return staffs[1:]
        if _is_close_to_image_top_or_bottom(
            staffs[-1], image
        ) and _have_all_the_same_number_of_staffs(staffs[:-1]):
            eprint("Removing last system from all voices, as it has a different number of staffs")
            return staffs[:-1]
    result: list[MultiStaff] = []
    for staff in staffs:
        result.extend(staff.break_apart())
    return sorted(result, key=lambda s: s.staffs[0].min_y)


def _get_number_of_voices(staffs: list[MultiStaff]) -> int:
    return len(staffs[0].staffs)


tr_omr_max_height = default_config.max_height
tr_omr_max_width = default_config.max_width


def get_tr_omr_canvas_size(
    image_shape: tuple[int, ...], margin_top: int = 0, margin_bottom: int = 0
) -> NDArray:
    tr_omr_max_height_with_margin = tr_omr_max_height - margin_top - margin_bottom
    tr_omr_ratio = float(tr_omr_max_height_with_margin) / tr_omr_max_width
    height, width = image_shape[:2]

    # Calculate the new size such that it fits exactly into the
    # tr_omr_max_height and tr_omr_max_width
    # while maintaining the aspect ratio of height and width.

    if height / width > tr_omr_ratio:
        # The height is the limiting factor.
        new_shape = [
            int(width / height * tr_omr_max_height_with_margin),
            tr_omr_max_height_with_margin,
        ]
    else:
        # The width is the limiting factor.
        new_shape = [tr_omr_max_width, int(height / width * tr_omr_max_width)]
    return np.array(new_shape)


def center_image_on_canvas(
    image: NDArray, canvas_size: NDArray, margin_top: int = 0, margin_bottom: int = 0
) -> NDArray:
    is_grayscale = image.ndim == 2 or (image.ndim == 3 and image.shape[2] == 1)

    resized = cv2.resize(image, canvas_size)  # type: ignore

    if is_grayscale:
        new_image = np.full(
            (tr_omr_max_height, tr_omr_max_width),
            255,
            dtype=np.uint8,
        )
    else:
        new_image = np.full(  # type: ignore
            (tr_omr_max_height, tr_omr_max_width, 3),
            255,
            dtype=np.uint8,
        )

    x_offset = 0
    tr_omr_max_height_with_margin = tr_omr_max_height - margin_top - margin_bottom
    y_offset = (tr_omr_max_height_with_margin - resized.shape[0]) // 2 + margin_top

    new_image[
        y_offset : y_offset + resized.shape[0],
        x_offset : x_offset + resized.shape[1],
    ] = resized

    return new_image


def add_image_into_tr_omr_canvas(image: NDArray) -> NDArray:
    new_shape = get_tr_omr_canvas_size(image.shape)
    new_image = center_image_on_canvas(image, new_shape)
    return new_image


def remove_black_contours_at_edges_of_image(gray: NDArray, unit_size: float) -> NDArray:
    _, thresh = cv2.threshold(gray, 97, 255, cv2.THRESH_BINARY)
    thresh = 255 - thresh
    contours, _hierarchy = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
    threshold = constants.black_spot_removal_threshold(unit_size)
    for cnt in contours:
        x, y, w, h = cv2.boundingRect(cnt)
        if w < threshold or h < threshold:
            continue
        is_at_edge_of_image = x == 0 or y == 0 or x + w == gray.shape[1] or y + h == gray.shape[0]
        if not is_at_edge_of_image:
            continue
        average_gray_intensity = 127
        is_mostly_dark = np.mean(thresh[y : y + h, x : x + w]) < average_gray_intensity  # type: ignore
        if is_mostly_dark:
            continue
        gray[y : y + h, x : x + w] = 255
    return gray


def _calculate_region(staff: Staff, regions: StaffRegions) -> NDArray:
    x_min = staff.min_x - 2 * staff.average_unit_size
    x_max = staff.max_x + 2 * staff.average_unit_size
    y_min = max(
        staff.min_y - 4 * staff.average_unit_size,
        regions.get_start_of_closest_staff_above(staff.min_y),
    )
    y_max = min(
        staff.max_y + 4 * staff.average_unit_size,
        regions.get_start_of_closest_staff_below(staff.max_y),
    )
    return np.array([int(x_min), int(y_min), int(x_max), int(y_max)])


def prepare_staff_image(
    debug: Debug, index: int, staff: Staff, staff_image: NDArray, regions: StaffRegions
) -> tuple[NDArray, Staff]:
    region = _calculate_region(staff, regions)
    image_dimensions = get_tr_omr_canvas_size(
        (int(region[3] - region[1]), int(region[2] - region[0]))
    )
    scaling_factor = image_dimensions[1] / (region[3] - region[1])
    staff_image = cv2.resize(
        staff_image,
        (int(staff_image.shape[1] * scaling_factor), int(staff_image.shape[0] * scaling_factor)),
    )
    region = np.round(region * scaling_factor)
    eprint("Dewarping staff", index)
    region_step1 = np.array(region) + np.array([-10, -50, 10, 50])
    staff_image, top_left = crop_image_and_return_new_top(staff_image, *region_step1)
    region_step2 = np.array(region) - np.array([*top_left, *top_left])
    top_left = top_left / scaling_factor
    staff = _dewarp_staff(staff, None, top_left, scaling_factor)
    dewarp = dewarp_staff_image(staff_image, staff, index, debug)
    staff_image = dewarp.dewarp(staff_image)
    staff_image, top_left = crop_image_and_return_new_top(staff_image, *region_step2)
    scaling_factor = 1

    eprint("Dewarping staff", index, "done")

    staff_image = remove_black_contours_at_edges_of_image(staff_image, staff.average_unit_size)
    staff_image = center_image_on_canvas(staff_image, image_dimensions)
    debug.write_image_with_fixed_suffix(f"_staff-{index}_input.jpg", staff_image)
    if debug.debug:
        transformed_staff = _dewarp_staff(staff, dewarp, top_left, scaling_factor)
        transformed_staff_image = staff_image.copy()
        for symbol in transformed_staff.symbols:
            center = symbol.center
            cv2.circle(transformed_staff_image, (int(center[0]), int(center[1])), 5, (0, 0, 255))
            cv2.putText(
                transformed_staff_image,
                type(symbol).__name__,
                (int(center[0]), int(center[1])),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.3,
                (0, 0, 255),
                1,
            )
        debug.write_image_with_fixed_suffix(
            f"_staff-{index}_debug_annotated.jpg", transformed_staff_image
        )
    return staff_image, staff


def _dewarp_staff(
    staff: Staff, dewarp: StaffDewarping | None, region: NDArray, scaling: float
) -> Staff:
    """
    Applies the same transformation on the staff coordinates as we did on the image.
    """

    def transform_coordinates(point: tuple[float, float]) -> tuple[float, float]:
        x, y = point
        x -= region[0]
        y -= region[1]
        if dewarp is not None:
            x, y = dewarp.dewarp_point((x, y))
        x = x * scaling
        y = y * scaling
        return x, y

    return staff.transform_coordinates(transform_coordinates)


def parse_staff_image(
    debug: Debug, index: int, staff: Staff, image: NDArray, regions: StaffRegions, config: Config
) -> list[EncodedSymbol]:
    staff_image, transformed_staff = prepare_staff_image(
        debug, index, staff, image, regions=regions
    )
    eprint("Running TrOmr inference on staff image", index)
    result = parse_staff_tromr(staff_image=staff_image, staff=transformed_staff, config=config)
    if debug.debug:
        result_image = staff_image.copy()
        for i, symbol in enumerate(result):
            center = symbol.coordinates
            if center is None or symbol.rhythm.startswith("chord"):
                continue
            if math.isnan(center[0]) or math.isnan(center[1]):
                continue
            center_int = (int(center[0]), int(center[1]))
            cv2.circle(result_image, center_int, 5, color=(0, 0, 255), thickness=2)
            cv2.putText(
                result_image,
                str(i) + ": " + symbol.rhythm,
                (center_int[0], center_int[1] - 10),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.3,
                (0, 0, 255),
                1,
            )

        debug.write_image_with_fixed_suffix(f"_staff-{index}_output.jpg", result_image)
    return result


def parse_staffs(
    debug: Debug, staffs: list[MultiStaff], image: NDArray, config: Config, selected_staff: int = -1
) -> list[list[EncodedSymbol]]:
    """
    Dewarps each staff and then runs it through an algorithm which extracts
    the rhythm and pitch information.
    """
    staffs = _ensure_same_number_of_staffs(staffs, image)
    # For simplicity we call every staff in a multi staff a voice,
    # even if it's part of a grand staff.
    number_of_voices = _get_number_of_voices(staffs)
    i = 0
    voices = []
    regions = StaffRegions(staffs)
    for voice in range(number_of_voices):
        staffs_for_voice = [staff.staffs[voice] for staff in staffs]
        result_for_voice = []
        for staff_index, staff in enumerate(staffs_for_voice):
            if selected_staff >= 0 and staff_index != selected_staff:
                eprint("Ignoring staff due to selected_staff argument", i)
                i += 1
                continue
            result_staff = parse_staff_image(debug, i, staff, image, regions, config)
            if len(result_staff) == 0:
                eprint("Skipping empty staff", i)
                i += 1
                continue
            result_staff.append(EncodedSymbol("newline"))
            result_for_voice.extend(result_staff)
            i += 1

        voices.append(remove_duplicated_symbols(result_for_voice))
    return voices



import numpy as np

from homr.type_definitions import NDArray


def find_peaks(
    x: NDArray,
    height: float | None = None,
    distance: float | None = None,
    prominence: float | None = None,
) -> tuple[NDArray, dict]:
    """
    Find peaks in a 1D array without using scipy.

    Parameters:
    -----------
    x : array-like
        1D signal data
    height : float, optional
        Minimum height of peaks
    distance : float, optional
        Minimum distance between peaks (in samples)
    prominence : float, optional
        Minimum prominence of peaks

    Returns:
    --------
    peaks : ndarray
        Indices of peaks
    properties : dict
        Dictionary with empty values (for compatibility)
    """
    x = np.asarray(x)
    if x.ndim != 1:
        raise ValueError("x must be 1D")
    if len(x) < 3:
        return np.array([], dtype=int), {}

    # Find local maxima, handling flat regions
    peaks_list: list[int] = []
    i = 1
    while i < len(x) - 1:
        # Check if we're at a local maximum or start of plateau
        if x[i] > x[i - 1]:
            # Find end of plateau
            j = i
            while j < len(x) - 1 and x[j] == x[j + 1]:
                j += 1

            # Check if plateau ends with a descent (making it a peak)
            if j < len(x) - 1 and x[j] > x[j + 1]:
                # Peak is at the midpoint of the plateau
                peak_idx = (i + j) // 2
                peaks_list.append(peak_idx)
                i = j + 1
            else:
                i = j + 1
        elif x[i] == x[i - 1]:
            # Start of a flat region from the left
            j = i
            while j < len(x) - 1 and x[j] == x[j + 1]:
                j += 1

            # Check if it's a peak (higher than left, and either at end or higher than right)
            if j < len(x) - 1 and x[j] > x[j + 1]:
                peak_idx = (i + j) // 2
                peaks_list.append(peak_idx)
            i = j + 1
        else:
            i += 1

    peaks = np.array(peaks_list, dtype=int)

    if len(peaks) == 0:
        return np.array([], dtype=int), {}

    # Apply height threshold
    if height is not None:
        mask = x[peaks] >= height
        peaks = peaks[mask]

    if len(peaks) == 0:
        return np.array([], dtype=int), {}

    # Apply prominence filter
    if prominence is not None:
        valid_peaks: list[int] = []
        for peak_val in peaks:
            peak = int(peak_val)
            # Calculate prominence properly
            # Find the lowest contour line around the peak
            # Look left for the minimum until we hit a higher peak or boundary
            left_min = x[peak]
            for k in range(peak - 1, -1, -1):
                if x[k] > x[peak]:
                    break
                left_min = min(left_min, x[k])

            # Look right for the minimum until we hit a higher peak or boundary
            right_min = x[peak]
            for k in range(peak + 1, len(x)):
                if x[k] > x[peak]:
                    break
                right_min = min(right_min, x[k])

            # Prominence is height above the higher of the two minima
            peak_prominence = x[peak] - max(left_min, right_min)

            if peak_prominence >= prominence:
                valid_peaks.append(peak)
        peaks = np.array(valid_peaks, dtype=int)

    if len(peaks) == 0:
        return np.array([], dtype=int), {}

    # Apply distance constraint (keep highest peaks)
    if distance is not None and len(peaks) > 1:
        # Sort peaks by height (descending)
        sorted_indices = np.argsort(x[peaks])[::-1]
        sorted_peaks = peaks[sorted_indices]

        keep: list[int] = []
        for peak_val in sorted_peaks:
            peak = int(peak_val)
            # Check if this peak is far enough from all kept peaks
            if len(keep) == 0 or all(abs(k - peak) >= distance for k in keep):
                keep.append(peak)

        # Sort back to original order
        peaks = np.array(sorted(keep), dtype=int)

    return peaks, {}



import glob
import os
from collections.abc import Sequence
from itertools import chain

import cv2

from homr.bounding_boxes import DebugDrawable
from homr.type_definitions import NDArray


class Debug:
    def __init__(self, original_image: NDArray, filename: str, debug: bool):
        self.filename = filename
        self.original_image = original_image
        filename = filename.replace("\\", "/")
        self.dir_name = os.path.dirname(filename)
        self.base_filename = os.path.join(self.dir_name, filename.split("/")[-1].split(".")[0])
        self.debug = debug
        self.colors = [
            (0, 255, 0),
            (0, 0, 255),
            (255, 0, 0),
            (255, 255, 0),
            (0, 255, 255),
            (255, 0, 255),
            (255, 165, 0),
            (255, 182, 193),
            (128, 0, 128),
            (64, 224, 208),
        ]
        self.debug_output_counter = 0
        self.written_files: list[str] = []

    def clean_debug_files_from_previous_runs(self) -> None:
        prefixes = (
            self.base_filename + "_debug_",
            self.base_filename + "_tesseract_input",
            self.base_filename + "_staff-",
        )

        for file in glob.glob(self.base_filename + "*"):
            if file.startswith(prefixes) and file not in self.written_files:
                os.remove(file)

    def _debug_file_name(self, suffix: str) -> str:
        self.debug_output_counter += 1
        return f"{self.base_filename}_debug_{str(self.debug_output_counter)}_{suffix}.png"

    def write_threshold_image(self, suffix: str, image: NDArray) -> None:
        if not self.debug:
            return
        filename = self._debug_file_name(suffix)
        self._remember_file_name(filename)
        cv2.imwrite(filename, 255 * image)

    def _remember_file_name(self, filename: str) -> None:
        self.written_files.append(filename)

    def write_bounding_boxes(self, suffix: str, bounding_boxes: Sequence[DebugDrawable]) -> None:
        if not self.debug:
            return
        img = self.original_image.copy()
        for box in bounding_boxes:
            box.draw_onto_image(img)
        filename = self._debug_file_name(suffix)
        self._remember_file_name(filename)
        cv2.imwrite(filename, img)

    def write_image(self, suffix: str, image: NDArray) -> None:
        if not self.debug:
            return
        filename = self._debug_file_name(suffix)
        self._remember_file_name(filename)
        cv2.imwrite(filename, image)

    def write_image_with_fixed_suffix(self, suffix: str, image: NDArray) -> None:
        if not self.debug:
            return
        filename = self.base_filename + suffix
        self._remember_file_name(filename)
        cv2.imwrite(filename, image)

    def write_all_bounding_boxes_alternating_colors(
        self, suffix: str, *boxes: Sequence[DebugDrawable]
    ) -> None:
        self.write_bounding_boxes_alternating_colors(suffix, list(chain.from_iterable(boxes)))

    def write_bounding_boxes_alternating_colors(
        self, suffix: str, bounding_boxes: Sequence[DebugDrawable]
    ) -> None:
        if not self.debug:
            return
        self.write_teaser(self._debug_file_name(suffix), bounding_boxes)

    def write_teaser(self, filename: str, bounding_boxes: Sequence[DebugDrawable]) -> None:
        img = self.original_image.copy()
        for i, box in enumerate(bounding_boxes):
            color = self.colors[i % len(self.colors)]
            box.draw_onto_image(img, color)
        self._remember_file_name(filename)
        cv2.imwrite(filename, img)

    def write_model_input_image(self, suffix: str, staff_image: NDArray) -> str:
        """
        These files aren't really debug files, but it's convenient to handle them here
        so that they are cleaned up together with the debug files.

        Model input images are the input to the transformer or OCR images.
        """
        filename = self.base_filename + suffix
        if self.debug:
            self._remember_file_name(filename)
        cv2.imwrite(filename, staff_image)
        return filename



import re
import threading
from concurrent.futures import Future, ThreadPoolExecutor

from rapidocr_onnxruntime import RapidOCR

from homr.debug import Debug
from homr.model import Staff

# Globals
_reader: RapidOCR | None = None
_reader_lock = threading.Lock()
_executor: ThreadPoolExecutor = ThreadPoolExecutor(max_workers=1)


def cleanup_text(text: str) -> str:
    """
    Remove all special characters from the text. Merge multiple whitespaces into a single space.
    """
    return re.sub(r"[^a-zA-Z0-9]+", " ", text).strip()


def _initialize_reader() -> None:
    """
    Thread-safe initialization of the OCR reader.
    """
    global _reader  # noqa: PLW0603
    if _reader is None:
        with _reader_lock:
            if _reader is None:  # double-checked locking
                _reader = RapidOCR()


def is_tempo_marking(text: str) -> bool:
    """
    Filter out tempo markings, e.g. 60 BPM
    """
    min_letters = 4
    if len(text) < min_letters:
        return True
    return sum(1 for c in text if "a" <= c.lower() <= "z") < min_letters


def _detect_title_task(debug: Debug, top_staff: Staff) -> str:
    _initialize_reader()

    if _reader is None:
        raise ValueError("reader is not initialized")
    image = debug.original_image
    height: int = int(15 * top_staff.average_unit_size)
    y: int = max(int(top_staff.min_y) - height, 0)
    x: int = max(int(top_staff.min_x) - 50, 0)
    width: int = int(top_staff.max_x - top_staff.min_x) + 100
    width = min(width, image.shape[1] - x)
    height = min(height, int(top_staff.min_y) - y)
    above_staff = image[y : y + height, x : x + width]

    ocr_input: str = debug.write_model_input_image("_tesseract_input.png", above_staff)
    ocr_results = _reader(ocr_input)

    if not ocr_results or not ocr_results[0]:
        return ""

    filtered_results = [r for r in ocr_results[0] if not is_tempo_marking(r[1])]

    if not filtered_results:
        return ""

    def bbox_height(bbox: list[list[float]]) -> float:
        ys = [p[1] for p in bbox]
        return max(ys) - min(ys)

    # Take max average character height = bbox_height / number of chars
    def font_size_score(result: tuple[list[list[float]], str, float]) -> float:
        bbox, text, _ = result
        if not text:
            return 0
        return bbox_height(bbox) / len(text)

    largest_text = max(filtered_results, key=font_size_score)[1]
    return cleanup_text(largest_text)


def download_ocr_weights() -> None:
    """
    Pre-download and initialize OCR reader weights before detection runs.
    """
    _initialize_reader()


def detect_title(debug: Debug, top_staff: Staff) -> Future[str]:
    """
    Runs the title detection in a separate thread and returns a Future.
    """
    return _executor.submit(_detect_title_task, debug, top_staff)



import cv2
import numpy as np

from homr import constants
from homr.debug import Debug
from homr.model import InputPredictions
from homr.simple_logging import eprint
from homr.type_definitions import NDArray


def estimate_noise(gray: NDArray) -> int:
    H, W = gray.shape
    M = np.array([[1, -2, 1], [-2, 4, -2], [1, -2, 1]])
    sigma = np.sum(np.sum(np.absolute(cv2.filter2D(gray, cv2.CV_64F, M)))) / (H * W)
    return sigma


def create_noise_grid(gray: NDArray, debug: Debug) -> NDArray | None:
    imgheight, imgwidth = gray.shape
    M, N = imgheight // 20, imgwidth // 20

    debug_image = cv2.cvtColor(gray, cv2.COLOR_GRAY2BGR)
    mask = np.zeros(gray.shape, dtype=np.uint8)

    grid = create_grid(gray, M, N)
    filtered_cells, total_cells = apply_noise_filter(grid, mask, debug_image, M, N)

    if debug.debug:
        debug.write_image("noise_crop", debug_image)

    return handle_filter_results(filtered_cells, total_cells, mask)


def create_grid(gray: NDArray, M: int, N: int) -> NDArray:
    imgheight, imgwidth = gray.shape
    grid = np.zeros([int(np.ceil(imgheight / M)), int(np.ceil(imgwidth / N))], dtype=np.uint8)

    for i, y1 in enumerate(range(0, imgheight, M)):
        for j, x1 in enumerate(range(0, imgwidth, N)):
            y2, x2 = y1 + M, x1 + N
            tile = gray[y1:y2, x1:x2]
            noise = estimate_noise(tile)
            grid[i, j] = noise

    return grid


def apply_noise_filter(
    grid: NDArray, mask: NDArray, debug_image: NDArray, M: int, N: int
) -> tuple[int, int]:
    imgheight, imgwidth = mask.shape
    filtered_cells, total_cells = 0, 0

    for i, y1 in enumerate(range(0, imgheight, M)):
        for j, x1 in enumerate(range(0, imgwidth, N)):
            y2, x2 = y1 + M, x1 + N
            noise = grid[i, j]
            neighbors = get_neighbors(grid, i, j)
            any_neighbor_above_limit = np.any(np.array(neighbors) > constants.image_noise_limit)

            if noise > constants.image_noise_limit and any_neighbor_above_limit:
                cv2.rectangle(debug_image, (x1, y1), (x2, y2), (0, 255, 255))
                filtered_cells += 1
            else:
                mask[y1:y2, x1:x2] = 255
                cv2.rectangle(debug_image, (x1, y1), (x2, y2), (0, 255, 0))

            cv2.putText(
                debug_image,
                f"{noise:.2f}",
                (x1 + N // 2, y1 + M // 2),
                cv2.FONT_HERSHEY_PLAIN,
                1,
                (0, 0, 255),
            )
            total_cells += 1

    return filtered_cells, total_cells


def get_neighbors(grid: NDArray, i: int, j: int) -> list[int]:
    neighbors = []
    if i > 0:
        neighbors.append(grid[i - 1, j])
    if j > 0:
        neighbors.append(grid[i, j - 1])
    if i < grid.shape[0] - 1:
        neighbors.append(grid[i + 1, j])
    if j < grid.shape[1] - 1:
        neighbors.append(grid[i, j + 1])
    return neighbors


def handle_filter_results(filtered_cells: int, total_cells: int, mask: NDArray) -> NDArray | None:
    half = 0.5
    if filtered_cells / total_cells > half:
        eprint(
            f"Would filter more than 50% of the image with {filtered_cells} of {total_cells} "
            + "cells, skipping noise filtering"
        )
        return None
    elif filtered_cells > 0:
        eprint(f"Filtered {filtered_cells} of {total_cells} cells")
        return mask
    return None


def filter_predictions(prediction: InputPredictions, debug: Debug) -> InputPredictions:
    mask = create_noise_grid(255 * prediction.staff, debug)
    if mask is None:
        return prediction
    return InputPredictions(
        original=cv2.bitwise_and(prediction.original, prediction.original, mask=mask),
        preprocessed=cv2.bitwise_and(prediction.preprocessed, prediction.preprocessed, mask=mask),
        notehead=cv2.bitwise_and(prediction.notehead, prediction.notehead, mask=mask),
        symbols=cv2.bitwise_and(prediction.symbols, prediction.symbols, mask=mask),
        staff=cv2.bitwise_and(prediction.staff, prediction.staff, mask=mask),
        clefs_keys=cv2.bitwise_and(prediction.clefs_keys, prediction.clefs_keys, mask=mask),
        stems_rest=cv2.bitwise_and(prediction.stems_rest, prediction.stems_rest, mask=mask),
    )



number_of_lines_on_a_staff = 5

max_number_of_ledger_lines = 4


def tolerance_for_staff_line_detection(unit_size: float) -> float:
    return unit_size / 3


def max_line_gap_size(unit_size: float) -> float:
    return 5 * unit_size


def is_short_line(unit_size: float) -> float:
    return unit_size / 5


def is_short_connected_line(unit_size: float) -> float:
    return 2 * unit_size


def min_height_for_brace_rough(unit_size: float) -> float:
    return 2 * unit_size


def max_width_for_brace_rough(unit_size: float) -> float:
    return 3 * unit_size


def min_height_for_brace(unit_size: float) -> float:
    return 4 * unit_size


def tolerance_for_touching_clefs(unit_size: float) -> int:
    return int(round(unit_size * 2))


def tolerance_for_staff_at_any_point(unit_size: float) -> int:
    return 0


def tolerance_note_grouping(unit_size: float) -> float:
    return 1 * unit_size


def bar_line_max_width(unit_size: float) -> float:
    return 2 * unit_size


def bar_line_min_height(unit_size: float) -> float:
    return 3 * unit_size


def black_spot_removal_threshold(unit_size: float) -> float:
    return 2 * unit_size


staff_line_segment_x_tolerance = 10

# We don't have to worried about mis-detections,
# because if not all staffs group the same way then we break the staffs up again
minimum_connections_to_form_combined_staff = 1

duration_of_quarter = 16

image_noise_limit = 50

staff_position_tolerance = 50

max_angle_for_lines_to_be_parallel = 10


NOTEHEAD_SIZE_RATIO = 1.285714  # width/height
